from abc import ABC, abstractmethod

_connector_table = {}


class Connector(ABC):

    def __init_subclass__(cls, connector_name, **kwargs):
        super().__init_subclass__(**kwargs)
        _connector_table[connector_name] = cls
        cls.extensions = {}

    @abstractmethod
    def renderC(self, scene, render_options, sensor_ids=[0], integrator_id=0):
        pass

    @abstractmethod
    def renderD(self, image_grads, scene, render_options, sensor_ids=[0], integrator_id=0):
        pass

    @classmethod
    def register(cls, class_name):
        def wrapper(func):
            cls.extensions[class_name] = func
            return func
        return wrapper

def is_connector_available(connector_name):
    return connector_name in _connector_table


def get_connector_list():
    return list(_connector_table.keys())


def get_connector(connector_name):
    return _connector_table[connector_name]()



from .io import to_torch_f
import torch 
import torch.nn.functional as F

def lookat(origin, target, up):
    origin = to_torch_f(origin)
    target = to_torch_f(target)
    up = to_torch_f(up)

    dir = F.normalize(target - origin, dim=0)
    left = F.normalize(torch.cross(up, dir), dim=0)
    new_up = F.normalize(torch.cross(dir, left), dim=0)

    to_world = to_torch_f(torch.eye(4))
    to_world[:3, 0] = left
    to_world[:3, 1] = new_up
    to_world[:3, 2] = dir
    to_world[:3, 3] = origin

    return to_world

def perspective(fov, aspect_ratio, near=1e-6, far=1e7):
    recip = 1 / (far - near)
    tan = torch.tan(torch.deg2rad(to_torch_f(fov * 0.5)))
    cot = 1 / tan

    mat = torch.diag(to_torch_f([cot, cot, far * recip, 0]))
    mat[2, 3] = -near * far * recip
    mat[3, 2] = 1

    mat = scale([-0.5, -0.5 * aspect_ratio, 1]) @ translate([-1, -1 / aspect_ratio, 0]) @ mat

    return mat

def perspective_full(fx, fy, cx, cy, aspect_ratio, near=1e-6, far=1e7):
    recip = 1 / (far - near)
    mat = torch.diag(to_torch_f([1, 1, far * recip, 0]))
    mat[2, 3] = -near * far * recip
    mat[3, 2] = 1

    mat = translate([1 - 2 * cx, 1 - 2 * cy, 0]) @ scale([2 * fx, 2 * fy, 1]) @ mat

    mat = scale([-0.5, -0.5 * aspect_ratio, 1]) @ translate([-1, -1 / aspect_ratio, 0]) @ mat
    mat = scale([-0.5, -0.5, 1]) @ translate([-1, -1, 0]) @ mat
    return mat

def batched_transform_pos(mat, vec):
    mat = mat.view(1, 4, 4)
    vec = vec.view(-1, 3, 1)
    tmp = (mat @ torch.cat([vec, torch.ones_like(vec)[:, 0:1]], dim=1)).reshape(-1, 4)
    return tmp[:, 0:3] / tmp[:, 3:]

def batched_transform_dir(mat, vec):
    mat = mat.view(1, 4, 4)
    vec = vec.view(-1, 3, 1)
    tmp = (mat @ torch.cat([vec, torch.zeros_like(vec)[:, 0:1]], dim=1)).reshape(-1, 4)
    return tmp[:, 0:3]

def translate(t_vec):
    t_vec = to_torch_f(t_vec)

    to_world = to_torch_f(torch.eye(4))
    to_world[:3, 3] = t_vec

    return to_world

def rotate(axis, angle, use_degree=True):
    axis = to_torch_f(axis)
    angle = to_torch_f(angle)

    to_world = to_torch_f(torch.eye(4))
    axis = F.normalize(axis, dim=0).reshape(3, 1)

    if use_degree:
        angle = torch.deg2rad(angle)

    sin_theta = torch.sin(angle)
    cos_theta = torch.cos(angle)

    cpm = to_torch_f(torch.zeros((3, 3)))
    cpm[0, 1] = -axis[2]
    cpm[0, 2] =  axis[1]
    cpm[1, 0] =  axis[2]
    cpm[1, 2] = -axis[0]
    cpm[2, 0] = -axis[1]
    cpm[2, 1] =  axis[0]

    R = cos_theta * to_torch_f(torch.eye(3))
    R += sin_theta * cpm
    R += (1 - cos_theta) * (axis @ axis.T)

    to_world[:3, :3] = R

    return to_world

def scale(size):
    size = to_torch_f(size)

    to_world = to_torch_f(torch.eye(4))

    if size.size() == () or size.size(dim=0) == 1:
        to_world[:3, :3] = to_torch_f(torch.eye(3)) * size
    elif size.size(dim=0) == 3:
        to_world[:3, :3] = torch.diag(size)
    else:
        print(f"unrecognized shape for size: {size.shape}")
        exit()

    return to_world

# texture map transform (2d)
def translate2D(t_vec):
    t_vec = to_torch_f(t_vec)

    to_world = to_torch_f(torch.eye(3))
    to_world[:2, 2] = t_vec

    return to_world

def rotate2D(angle, use_degree=True):
    angle = to_torch_f(angle)

    to_world = to_torch_f(torch.eye(3))

    if use_degree:
        angle = torch.deg2rad(angle)

    sin_theta = torch.sin(angle)
    cos_theta = torch.cos(angle)

    R = cos_theta * to_torch_f(torch.eye(2))

    R[0 ,1] = -sin_theta
    R[1 ,0] = sin_theta

    to_world[:2, :2] = R

    return to_world

def scale2D(size):
    size = to_torch_f(size)
    to_world = to_torch_f(torch.eye(3))

    if size.size(dim=0) == 1:
        to_world[:2, :2] = torch.diag(size) * to_torch_f(torch.eye(2))
    elif size.size(dim=0) == 2:
        to_world[:2, :2] = torch.diag(size)
    else:
        print(f"unrecognized shape for size: {size.shape}")
        exit()

    return to_world


import torch 

def sample_sphere(batch, radius, method='stratified', axis=1):
	# assumes y-axis is up by default, otherwise we swap 
	if method == 'uniform':
		phi = torch.rand(batch, 1) * torch.pi 
		theta = torch.rand(batch, 1) * 2 * torch.pi     
		sinPhi = torch.sin(phi)  
		cosPhi = torch.cos(phi) 
		sinTheta = torch.sin(theta) 
		cosTheta = torch.cos(theta) 
		samples = torch.cat([
			sinPhi * cosTheta, 
			cosPhi,
			sinPhi * sinTheta
		], dim=1) * radius

	elif method == 'stratified':
		cosPhi = torch.linspace(1.0 - 0.01, -1.0 + 0.01, batch).unsqueeze(-1)
		theta  = torch.linspace(0, torch.pi * 10, batch).unsqueeze(-1) \
			+ torch.rand(batch, 1) * 0.01
		
		sinPhi = torch.sqrt(1 - cosPhi * cosPhi)
		sinTheta = torch.sin(theta) 
		cosTheta = torch.cos(theta) 
		samples = torch.cat([
			sinPhi * cosTheta, 
			cosPhi,
			sinPhi * sinTheta
		], dim=1) * radius
		
	elif method == 'fibonacci':
		# From http://extremelearning.com.au/how-to-evenly-distribute-points-on-a-sphere-more-effectively-than-the-canonical-fibonacci-lattice/
		golden_ratio = (1 + 5**0.5) / 2
		i = torch.arange(batch).unsqueeze(-1)
		theta = 2 * torch.pi * i / golden_ratio
		phi = torch.acos(1 - 2 * (i + 0.5) / batch)
		sinPhi = torch.sin(phi)  
		cosPhi = torch.cos(phi)
		sinTheta = torch.sin(theta) 
		cosTheta = torch.cos(theta) 
		samples = torch.cat([
			sinPhi * cosTheta, 
			cosPhi,
			sinPhi * sinTheta
		], dim=1) * radius
	
	# Switch axis 
	index = torch.LongTensor([0, 1, 2])
	index[axis] = 1 
	index[1] = axis 
	points = samples[:, index]

	return points

def sample_hemisphere(batch, radius, method='stratified', axis=1):
	# assumes y-axis is up by default, otherwise we swap 
	if method == 'uniform':
		phi = torch.rand(batch, 1) * torch.pi * 0.5
		theta = torch.rand(batch, 1) * 2 * torch.pi     
		sinPhi = torch.sin(phi)  
		cosPhi = torch.cos(phi) 
		sinTheta = torch.sin(theta) 
		cosTheta = torch.cos(theta) 
		samples = torch.cat([
			sinPhi * cosTheta, 
			cosPhi,
			sinPhi * sinTheta
		], dim=1) * radius

	elif method == 'stratified':
		cosPhi = torch.linspace(1.0 - 0.01, 0.0 + 0.01, batch).unsqueeze(-1)
		theta  = torch.linspace(0, torch.pi * 10, batch).unsqueeze(-1) \
			+ torch.rand(batch, 1) * 0.01
		
		sinPhi = torch.sqrt(1 - cosPhi * cosPhi)
		sinTheta = torch.sin(theta) 
		cosTheta = torch.cos(theta) 
		samples = torch.cat([
			sinPhi * cosTheta, 
			cosPhi,
			sinPhi * sinTheta
		], dim=1) * radius
		
	elif method == 'fibonacci':
		# From http://extremelearning.com.au/how-to-evenly-distribute-points-on-a-sphere-more-effectively-than-the-canonical-fibonacci-lattice/
		golden_ratio = (1 + 5**0.5) / 2
		i = torch.arange(batch).unsqueeze(-1)
		theta = 2 * torch.pi * i / golden_ratio
		phi = torch.acos(1 - (i + 0.5) / batch)
		sinPhi = torch.sin(phi)  
		cosPhi = torch.cos(phi)
		sinTheta = torch.sin(theta) 
		cosTheta = torch.cos(theta) 
		samples = torch.cat([
			sinPhi * cosTheta, 
			cosPhi,
			sinPhi * sinTheta
		], dim=1) * radius
	
	# Switch axis 
	index = torch.LongTensor([0, 1, 2])
	index[axis] = 1 
	index[1] = axis 
	points = samples[:, index]

	return points


from ..connector import Connector
from ..scene import *
from ..io import write_mesh
from collections import OrderedDict

import drjit
import psdr_jit
from drjit.scalar import Array3f
from drjit.cuda import Array3f as Vector3fC, Array3i as Vector3iC
from drjit.cuda.ad import Array3f as Vector3fD, Array1f as Vector1fD, Float32 as FloatD, Matrix4f as Matrix4fD, Matrix3f as Matrix3fD
from drjit.cuda.ad import Float32 as FloatD
import torch

import os

class PSDRJITConnector(Connector, connector_name='psdr_jit'):

    def __init__(self):
        super().__init__()

    def update_scene_objects(self, scene, render_options):
        if 'psdr_jit' in scene.cached:
            cache = scene.cached['psdr_jit']
        else:
            cache = {}
            scene.cached['psdr_jit'] = cache
            cache['scene'] = psdr_jit.Scene()
            cache['name_map'] = {}
            cache['integrators'] = OrderedDict()
            cache['configured'] = False

            def clean_up():
                drjit.flush_malloc_cache()
                drjit.flush_kernel_cache()

            cache['clean_up'] = clean_up

        cache['scene'].opts.spp = render_options['spp']
        cache['scene'].opts.sppe = render_options['sppe']
        cache['scene'].opts.sppse = render_options['sppse']
        cache['scene'].opts.log_level = render_options['log_level']

        drjit_params = []
        for name in scene.components:
            component = scene[name]
            drjit_params += self.extensions[type(component)](name, scene)

        if not cache['configured']:
            cache['scene'].configure() 
            cache['configured'] = True
        
        return scene.cached['psdr_jit'], drjit_params

    def renderC(self, scene, render_options, sensor_ids=[0], integrator_id=0):
        cache, _ = self.update_scene_objects(scene, render_options)

        cache['scene'].configure(sensor_ids)

        npass = render_options['npass']
        h, w, c = cache['film']['shape']
        if type(integrator_id) == int:
            integrator = list(cache['integrators'].values())[integrator_id]
        elif type(integrator_id) == str:
            integrator = cache['integrators'][integrator_id]
        else:
            raise RuntimeError('integrator_id is invalid: {integrator_id}')

        images = []
        for sensor_id in sensor_ids:
            image = to_torch_f(torch.zeros((h * w, c)))
            for i in range(npass):
                image_pass = integrator.renderC(cache['scene'], sensor_id).torch()
                image += image_pass / npass
            image = image.reshape(h, w, c)
            images.append(image)

        return images
        
    def renderD(self, image_grads, scene, render_options, sensor_ids=[0], integrator_id=0):
        cache, drjit_params = self.update_scene_objects(scene, render_options)

        cache['scene'].configure(sensor_ids)

        npass = render_options['npass']
        h, w, c = cache['film']['shape']
        if type(integrator_id) == int:
            integrator = list(cache['integrators'].values())[integrator_id]
        elif type(integrator_id) == str:
            integrator = cache['integrators'][integrator_id]
        else:
            raise RuntimeError('integrator_id is invalid: {integrator_id}')
        
        param_grads = [torch.zeros_like(scene[param_name]) for param_name in scene.requiring_grad]

        for i, sensor_id in enumerate(sensor_ids):
            image_grad = Vector3fC(image_grads[i].reshape(-1, 3) / npass)
            for j in range(npass):
                image = integrator.renderD(cache['scene'], sensor_id)
                tmp = drjit.dot(image_grad, image)
                drjit.backward(tmp)

                for param_grad, drjit_param in zip(param_grads, drjit_params):
                    grad = to_torch_f(drjit.grad(drjit_param).torch())
                    grad = torch.nan_to_num(grad).reshape(param_grad.shape)
                    param_grad += grad

        return param_grads       
    
    def forward_ad_mesh_translation(self, mesh_id, scene, render_options, sensor_ids=[0], integrator_id=0):
        cache, drjit_params = self.update_scene_objects(scene, render_options)
        assert len(drjit_params) == 0

        P = FloatD(0.) 
        drjit.enable_grad(P) 
        psdr_mesh = cache['scene'].param_map[cache['name_map'][mesh_id]]
        psdr_mesh.set_transform(Matrix4fD([[1.,0.,0.,P],[0.,1.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.],]))

        cache['scene'].configure(sensor_ids)

        npass = render_options['npass']
        h, w, c = cache['film']['shape']
        if type(integrator_id) == int:
            integrator = list(cache['integrators'].values())[integrator_id]
        elif type(integrator_id) == str:
            integrator = cache['integrators'][integrator_id]
        else:
            raise RuntimeError('integrator_id is invalid: {integrator_id}')

        
        image = to_torch_f(torch.zeros((h * w, c)))
        grad_image = to_torch_f(torch.zeros((h * w, c)))

        for j in range(npass):
            drjit_image = integrator.renderD(cache['scene'], sensor_ids[0])
            image += to_torch_f(drjit_image.torch()) / npass

            drjit.set_grad(P, 1.0)
            drjit.forward_to(drjit_image)
            drjit_grad_image = drjit.grad(drjit_image)
            grad_image += to_torch_f(drjit_grad_image.torch()) / npass

        image = image.reshape(h, w, c)
        grad_image = grad_image.reshape(h, w, c)
        return image, grad_image

@PSDRJITConnector.register(Integrator)
def process_integrator(name, scene):
    integrator = scene[name]
    cache = scene.cached['psdr_jit']

    if integrator['type'] == 'field':
        psdr_integrator = psdr_jit.FieldExtractionIntegrator(integrator['config']['type'])
        cache['integrators'][name] = psdr_integrator
    elif integrator['type'] == 'collocated':
        psdr_integrator = psdr_jit.CollocatedIntegrator(integrator['config']['intensity'])
        cache['integrators'][name] = psdr_integrator
    elif integrator['type'] == 'path':
        psdr_integrator = psdr_jit.PathTracer(integrator['config']['max_depth'])
        cache['integrators'][name] = psdr_integrator
        if 'hide_emitters' in integrator['config']:
            psdr_integrator.hide_emitters = integrator['config']['hide_emitters']
    else:
        raise RuntimeError(f"unrecognized integrator type: {integrator['type']}")

    return []

@PSDRJITConnector.register(HDRFilm)
def process_hdr_film(name, scene):
    film = scene[name]
    cache = scene.cached['psdr_jit']

    cache['film'] = {
        'shape': (film['height'], film['width'], 3)
    }
    cache['scene'].opts.width = film['width']
    cache['scene'].opts.height = film['height']

    return []

@PSDRJITConnector.register(PerspectiveCamera)
def process_perspective_camera(name, scene):
    sensor = scene[name]
    cache = scene.cached['psdr_jit']
    psdr_scene = cache['scene']

    # Create the object if it has not been created
    if name not in cache['name_map']:
        psdr_sensor = psdr_jit.PerspectiveCamera(sensor['fov'], sensor['near'], sensor['far'])
        psdr_sensor.to_world = Matrix4fD(sensor['to_world'].reshape(1, 4, 4))
        psdr_scene.add_Sensor(psdr_sensor)
        cache['name_map'][name] = f"Sensor[{psdr_scene.num_sensors - 1}]"

    psdr_sensor = psdr_scene.param_map[cache['name_map'][name]]
    
    # Update parameters
    updated = sensor.get_updated()
    if len(updated) > 0:
        for param_name in updated:
            if param_name == "to_world":
                psdr_sensor.to_world = Matrix4fD(sensor['to_world'].reshape(1, 4, 4))
            sensor.params[param_name]['updated'] = False

    # Enable grad for parameters requiring grad
    drjit_params = []
    requiring_grad = sensor.get_requiring_grad()
    if len(requiring_grad) > 0:
        for param_name in requiring_grad:
            if param_name == "to_world":
                drjit_param = psdr_sensor.to_world
                drjit.enable_grad(drjit_param)
                drjit_params.append(drjit_param)

    return drjit_params

@PSDRJITConnector.register(PerspectiveCameraFull)
def process_perspective_camera_full(name, scene):
    sensor = scene[name]
    cache = scene.cached['psdr_jit']
    psdr_scene = cache['scene']

    # Create the object if it has not been created
    if name not in cache['name_map']:
        psdr_sensor = psdr_jit.PerspectiveCamera(sensor['fx'], sensor['fy'], sensor['cx'], sensor['cy'], sensor['near'], sensor['far'])
        psdr_sensor.to_world = Matrix4fD(sensor['to_world'].reshape(1, 4, 4))
        psdr_scene.add_Sensor(psdr_sensor)
        cache['name_map'][name] = f"Sensor[{psdr_scene.num_sensors - 1}]"

    psdr_sensor = psdr_scene.param_map[cache['name_map'][name]]
    
    # Update parameters
    updated = sensor.get_updated()
    if len(updated) > 0:
        for param_name in updated:
            if param_name == "to_world":
                psdr_sensor.to_world = Matrix4fD(sensor['to_world'].reshape(1, 4, 4))
            sensor.params[param_name]['updated'] = False

    # Enable grad for parameters requiring grad
    drjit_params = []
    requiring_grad = sensor.get_requiring_grad()
    if len(requiring_grad) > 0:
        for param_name in requiring_grad:
            if param_name == "to_world":
                drjit_param = psdr_sensor.to_world
                drjit.enable_grad(drjit_param)
                drjit_params.append(drjit_param)

    return drjit_params

@PSDRJITConnector.register(Mesh)
def process_mesh(name, scene):
    mesh = scene[name]
    cache = scene.cached['psdr_jit']
    psdr_scene = cache['scene']

    # Create the object if it has not been created
    if name not in cache['name_map']:
        # Create its material first
        mat_id = mesh['mat_id']
        if mat_id not in scene:
            raise RuntimeError(f"The material of the mesh {name} doesn't exist: mat_id={mat_id}")
        brdf = scene[mat_id]
        PSDRJITConnector.extensions[type(brdf)](mat_id, scene)

        # TODO: Fix this workaround when psdr-jit updates psdr_mesh.load_raw()
        if mesh['can_change_topology']:
            psdr_mesh = psdr_jit.Mesh()
            psdr_mesh.load_raw(Vector3fC(mesh['v']), Vector3iC(mesh['f']))
            psdr_mesh.use_face_normal = mesh['use_face_normal']

            psdr_emitter = psdr_jit.AreaLight(mesh['radiance'].tolist()) if mesh['is_emitter'] else None
            psdr_scene.add_Mesh(psdr_mesh, mat_id, psdr_emitter)
        else:
            write_mesh('__psdr_jit_tmp__.obj', mesh['v'], mesh['f'], mesh['uv'], mesh['fuv'])
            psdr_emitter = psdr_jit.AreaLight(mesh['radiance'].tolist()) if mesh['is_emitter'] else None
            psdr_scene.add_Mesh('__psdr_jit_tmp__.obj', mesh['to_world'].reshape(1, 4, 4), mat_id, psdr_emitter)
            os.remove('__psdr_jit_tmp__.obj')
        
        cache['name_map'][name] = f"Mesh[{psdr_scene.num_meshes - 1}]"

    psdr_mesh = psdr_scene.param_map[cache['name_map'][name]]

    # Update parameters
    updated = mesh.get_updated()
    if len(updated) > 0:
        for param_name in updated:
            if param_name == 'v':
                if mesh['can_change_topology']:
                    psdr_mesh.load_raw(Vector3fC(mesh['v']), Vector3iC(mesh['f']))
                else:
                    psdr_mesh.vertex_positions = Vector3fC(mesh['v'])
                    psdr_mesh.face_indices = Vector3iC(mesh['f'])

            mesh.params[param_name]['updated'] = False

    # Enable grad for parameters requiring grad
    drjit_params = []
    requiring_grad = mesh.get_requiring_grad()
    if len(requiring_grad) > 0:
        for param_name in requiring_grad:
            if param_name == 'v':
                drjit_param = psdr_mesh.vertex_positions
                drjit.enable_grad(drjit_param)
                drjit_params.append(drjit_param)

    return drjit_params

def convert_color(color, c, bitmap=True):
    if color.shape == ():
        color = color.tile(c)
    if color.shape == (c,):
        color = color.reshape(1, 1, c)
    h, w, _ = color.shape
    if c == 3:
        color = Vector3fD(color.reshape(-1, c))
        if bitmap:
            color = psdr_jit.Bitmap3fD(w, h, color)
    elif c == 1:
        color = FloatD(color.reshape(-1))
        if bitmap:
            color = psdr_jit.Bitmap1fD(w, h, color)
    else:
        raise RuntimeError("Not support bitmap channel number: {c}")
    return color

@PSDRJITConnector.register(DiffuseBRDF)
def process_diffuse_brdf(name, scene):
    brdf = scene[name]
    cache = scene.cached['psdr_jit']
    psdr_scene = cache['scene']
    
    # Create the object if it has not been created
    if name not in cache['name_map']:
        d = convert_color(brdf['d'], 3)
        psdr_bsdf = psdr_jit.DiffuseBSDF(d)
        psdr_scene.add_BSDF(psdr_bsdf, name)
        cache['name_map'][name] = f"BSDF[id={name}]"

    psdr_brdf = psdr_scene.param_map[cache['name_map'][name]]

    # Update parameters
    updated = brdf.get_updated()
    if len(updated) > 0:
        for param_name in updated:
            if param_name == 'd':
                psdr_brdf.reflectance = convert_color(brdf['d'], 3)
            brdf.params[param_name]['updated'] = False

    # Enable grad for parameters requiring grad
    drjit_params = []
    requiring_grad = brdf.get_requiring_grad()
    if len(requiring_grad) > 0:
        for param_name in requiring_grad:
            if param_name == 'd':
                drjit_param = psdr_brdf.reflectance.data
                drjit.enable_grad(drjit_param)
                drjit_params.append(drjit_param)

    return drjit_params

@PSDRJITConnector.register(MicrofacetBRDF)
def process_microfacet_brdf(name, scene):
    brdf = scene[name]
    cache = scene.cached['psdr_jit']
    psdr_scene = cache['scene']
    
    # Create the object if it has not been created
    if name not in cache['name_map']:
        d = convert_color(brdf['d'], 3)
        s = convert_color(brdf['s'], 3)
        r = convert_color(brdf['r'], 1)

        psdr_bsdf = psdr_jit.MicrofacetBSDF(s, d, r)
        psdr_scene.add_BSDF(psdr_bsdf, name)
        cache['name_map'][name] = f"BSDF[id={name}]"

    psdr_brdf = psdr_scene.param_map[cache['name_map'][name]]

    # Update parameters
    updated = brdf.get_updated()
    if len(updated) > 0:
        for param_name in updated:
            if param_name == 'd':
                psdr_brdf.diffuseReflectance.data = convert_color(brdf['d'], 3, bitmap=False)
            elif param_name == 's':
                psdr_brdf.specularReflectance.data = convert_color(brdf['s'], 3, bitmap=False)
            elif param_name == 'r':
                psdr_brdf.roughness.data = convert_color(brdf['r'], 1, bitmap=False)
            brdf.params[param_name]['updated'] = False

    # Enable grad for parameters requiring grad
    drjit_params = []
    
    def enable_grad(drjit_param):
        drjit.enable_grad(drjit_param)
        drjit_params.append(drjit_param)

    requiring_grad = brdf.get_requiring_grad()
    if len(requiring_grad) > 0:
        for param_name in requiring_grad:
            if param_name == 'd':
                enable_grad(psdr_brdf.diffuseReflectance.data)
            elif param_name == 's':
                enable_grad(psdr_brdf.specularReflectance.data)
            elif param_name == 'r':
                enable_grad(psdr_brdf.roughness.data)

    return drjit_params

@PSDRJITConnector.register(EnvironmentLight)
def process_environment_light(name, scene):
    emitter = scene[name]
    cache = scene.cached['psdr_jit']
    psdr_scene = cache['scene']
    
    # Create the object if it has not been created
    if name not in cache['name_map']:
        radiance = convert_color(emitter['radiance'], 3)

        psdr_emitter = psdr_jit.EnvironmentMap()
        psdr_emitter.radiance = radiance
        psdr_emitter.set_transform(Matrix4fD(emitter['to_world'].reshape(1, 4, 4)))
        psdr_scene.add_EnvironmentMap(psdr_emitter)
        cache['name_map'][name] = f"Emitter[{psdr_scene.get_num_emitters() - 1}]"

    psdr_emitter = psdr_scene.param_map[cache['name_map'][name]]

    # Update parameters
    updated = emitter.get_updated()
    if len(updated) > 0:
        for param_name in updated:
            if param_name == 'radiance':
                psdr_emitter.radiance = convert_color(emitter['radiance'], 3)
            elif param_name == 'to_world':
                psdr_emitter.set_transform(Matrix4fD(emitter['to_world'].reshape(1, 4, 4)))
            emitter.params[param_name]['updated'] = False

    # Enable grad for parameters requiring grad
    drjit_params = []
    
    def enable_grad(drjit_param):
        drjit.enable_grad(drjit_param)
        drjit_params.append(drjit_param)

    requiring_grad = emitter.get_requiring_grad()
    if len(requiring_grad) > 0:
        for param_name in requiring_grad:
            if param_name == 'radiance':
                enable_grad(psdr_emitter.radiance.data)
                
    return drjit_params


# Scene components specfic to psdr-jit
class MicrofacetBRDFPerVertex(ParamGroup):

    def __init__(self, d, s, r):
        super().__init__()
        
        self.add_param('d', to_torch_f(d), is_tensor=True, is_diff=True, help_msg='diffuse reflectance')
        self.add_param('s', to_torch_f(s), is_tensor=True, is_diff=True, help_msg='specular reflectance')
        self.add_param('r', to_torch_f(r), is_tensor=True, is_diff=True, help_msg='roughness')

@PSDRJITConnector.register(MicrofacetBRDFPerVertex)
def process_microfacet_brdf_per_vertex(name, scene):
    brdf = scene[name]
    cache = scene.cached['psdr_jit']
    psdr_scene = cache['scene']
    
    # Create the object if it has not been created
    if name not in cache['name_map']:
        d = Vector3fD(brdf['d'])
        s = Vector3fD(brdf['s'])
        r = Vector1fD(brdf['r'])

        psdr_bsdf = psdr_jit.MicrofacetBSDFPerVertex(s, d, r)
        psdr_scene.add_BSDF(psdr_bsdf, name)
        cache['name_map'][name] = f"BSDF[id={name}]"

    psdr_brdf = psdr_scene.param_map[cache['name_map'][name]]

    # Update parameters
    updated = brdf.get_updated()
    if len(updated) > 0:
        for param_name in updated:
            if param_name == 'd':
                psdr_brdf.diffuseReflectance = Vector3fD(brdf['d'])
            elif param_name == 's':
                psdr_brdf.specularReflectance = Vector3fD(brdf['s'])
            elif param_name == 'r':
                psdr_brdf.roughness= Vector1fD(brdf['r'])
            brdf.params[param_name]['updated'] = False

    # Enable grad for parameters requiring grad
    drjit_params = []
    
    def enable_grad(drjit_param):
        drjit.enable_grad(drjit_param)
        drjit_params.append(drjit_param)

    requiring_grad = brdf.get_requiring_grad()
    if len(requiring_grad) > 0:
        for param_name in requiring_grad:
            if param_name == 'd':
                enable_grad(psdr_brdf.diffuseReflectance)
            elif param_name == 's':
                enable_grad(psdr_brdf.specularReflectance)
            elif param_name == 'r':
                enable_grad(psdr_brdf.roughness)

    return drjit_params


import torch

def l1_loss(target_image, rendered_image):
    return (target_image - rendered_image).abs().mean() # NOTE

def mesh_laplacian_smoothing(verts, faces):
    # compute L once per mesh subdiv.
    with torch.no_grad():
        V, F = verts.shape[0], faces.shape[0]
        face_verts = verts[faces]
        v0, v1, v2 = face_verts[:, 0], face_verts[:, 1], face_verts[:, 2]
        # Side lengths of each triangle, of shape (sum(F_n),)
        # A is the side opposite v1, B is opposite v2, and C is opposite v3
        A = (v1 - v2).norm(dim=1)
        B = (v0 - v2).norm(dim=1)
        C = (v0 - v1).norm(dim=1)
        # Area of each triangle (with Heron's formula); shape is (sum(F_n),)
        s = 0.5 * (A + B + C)
        # note that the area can be negative (close to 0) causing nans after sqrt()
        # we clip it to a small positive value
        area = (s * (s - A) * (s - B) * (s - C)).clamp_(min=1e-6).sqrt()
        # Compute cotangents of angles, of shape (sum(F_n), 3)
        A2, B2, C2 = A * A, B * B, C * C
        cota = (B2 + C2 - A2) / area
        cotb = (A2 + C2 - B2) / area
        cotc = (A2 + B2 - C2) / area
        cot = torch.stack([cota, cotb, cotc], dim=1)
        cot /= 4.0
        # Construct a sparse matrix by basically doing:
        # L[v1, v2] = cota
        # L[v2, v0] = cotb
        # L[v0, v1] = cotc
        ii = faces[:, [1, 2, 0]]
        jj = faces[:, [2, 0, 1]]
        idx = torch.stack([ii, jj], dim=0).view(2, F * 3)
        L = torch.sparse.FloatTensor(idx, cot.view(-1), (V, V))
        # Make it symmetric; this means we are also setting
        # L[v2, v1] = cota
        # L[v0, v2] = cotb
        # L[v1, v0] = cotc
        L += L.t()
        norm_w = torch.sparse.sum(L, dim=1).to_dense().view(-1, 1)
        idx = norm_w > 1e-6
        norm_w[idx] = 1.0 / norm_w[idx]
    loss = L.mm(verts) * norm_w - verts
    loss = loss.norm(dim=1)
    return loss.mean()

def total_variation_loss(texture, texture_mask):
    m = (texture_mask[:-1, :-1] & texture_mask[1:, :-1] & texture_mask[:-1, 1:])
    loss = (texture[:-1, :-1] - texture[1:, :-1])[m].abs().mean() +\
           (texture[:-1, :-1] - texture[:-1, 1:])[m].abs().mean()
    return loss


from .io import to_numpy

import time
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import LinearSegmentedColormap, Normalize

class Timer:
    def __init__(self, label):
        self.label = label

    def __enter__(self):
        self.start_time = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        elapsed_time = time.time() - self.start_time
        print(f"[{self.label}] Elapsed time: {elapsed_time} seconds")

def apply_pmkmp_cm(image, vmin=0, vmax=1):
    """
    Apply the pmkmp color map to an image.
    The image should be an np.array of shape [H, W, 3].
    """
    gray_cm = plt.get_cmap('gray')

    cubicL_path = Path(__file__).parent / 'data' / 'cubicL.txt' 
    cubicL = np.loadtxt(cubicL_path)
    pmkmp_cm = LinearSegmentedColormap.from_list("cubicL", cubicL, N=256)

    norm = Normalize(vmin, vmax)
    image = norm(to_numpy(image))

    gray_image = gray_cm(image)[:, :, 0, 0]
    pmkmp_image = pmkmp_cm(gray_image)

    return pmkmp_image

def get_pmkmp_color_bar(cb_path='colorbar.png'):
    """
    Save an example pmkmp color bar to cb_path.
    """
    cubicL_path = Path(__file__).parent / 'data' / 'cubicL.txt' 
    cubicL = np.loadtxt(cubicL_path)
    cm = LinearSegmentedColormap.from_list("cubicL", cubicL, N=256)
    fig = plt.figure()
    ax = fig.add_axes([0.05, 0.80, 0.9, 0.1])
    cb = ColorbarBase(ax, orientation='horizontal', 
                                cmap=cm)
    plt.savefig(cb_path, bbox_inches='tight')


from .parameter import ParamGroup
from .transform import lookat, perspective, perspective_full, batched_transform_pos, batched_transform_dir
from .io import read_image, read_mesh, to_torch_f, to_torch_i

from collections import OrderedDict

import torch
import torch.nn.functional as F

class Scene:

    def __init__(self) -> None:
        self.components = OrderedDict()
        self.requiring_grad = ()
        self.cached = {}

    def set(self, name, component):
        self.components[name] = component

    def __getitem__(self, name):
        item = self.components
        for c in name.split('.'):
            item = item[c]
        return item
    
    def __setitem__(self, name, param_value):
        component_name, param_name = name.rsplit('.', 1)
        self[component_name][param_name] = param_value
    
    def __contains__(self, name):
        item = self.components
        for c in name.split('.'):
            if c not in item: return False
            item = item[c]
        return True

    def configure(self):
        requiring_grad = []
        for cname in self.components:
            requiring_grad += [f'{cname}.{pname}' for pname in self.components[cname].get_requiring_grad()]
        self.requiring_grad = tuple(requiring_grad)

    def __str__(self):
        lines = []

        for name in self.components:
            lines.append(f"{name}:")
            lines.append(str(self.components[name]))
            lines.append("\n")

        return '\n'.join(lines)
    
    def clear_cache(self):
        for connector in self.cached.keys():
            if 'clean_up' in self.cached[connector]:
                self.cached[connector]['clean_up']()
        self.cached = {}
        # Detach the tensors requiring grad
        for param_name in self.requiring_grad:
            self[param_name] = self[param_name].detach()

    def filter(self, component_type):
        return [cname for cname in self.components if component_type == type(self.components[cname])]

class Integrator(ParamGroup):

    def __init__(self, type, config):
        super().__init__()
        
        self.add_param('type', type, help_msg='integrator type')
        self.add_param('config', config, help_msg='integrator config')

class HDRFilm(ParamGroup):

    def __init__(self, width, height):
        super().__init__()
        
        self.add_param('width', width, help_msg='film width')
        self.add_param('height', height, help_msg='film height')

class PerspectiveCamera(ParamGroup):
    
    def __init__(self, fov, to_world, near=1e-6, far=1e7):
        super().__init__()

        self.add_param('fov', fov, help_msg='sensor fov')
        self.add_param('near', near, help_msg='sensor near clip')
        self.add_param('far', far, help_msg='sensor far clip')
        self.add_param('to_world', to_torch_f(to_world), is_tensor=True, is_diff=True, help_msg='sensor to_world matrix')

    def get_rays(self, samples, aspect_ratio):
        samples = torch.cat([samples, torch.zeros_like(samples)[:, 0:1]], dim=1)
        sample_to_camera = torch.inverse(perspective(self['fov'], aspect_ratio, self['near'], self['far']))
        rays_o = batched_transform_pos(self['to_world'], to_torch_f([[0, 0, 0]]))
        rays_d = F.normalize(batched_transform_pos(sample_to_camera, samples), dim=1)
        rays_d = batched_transform_dir(self['to_world'], rays_d)
        return rays_o.repeat(samples.shape[0], 1), rays_d

    @classmethod
    def from_lookat(cls, fov, origin, target, up, near=1e-6, far=1e7):
        sensor = cls(fov, torch.eye(4), near, far)
        origin = to_torch_f(origin)
        target = to_torch_f(target)
        up = to_torch_f(up)
        sensor['to_world'] = lookat(origin, target, up)
        return sensor

# To be merged with PerspectiveCamera
class PerspectiveCameraFull(ParamGroup):
    
    def __init__(self, fx, fy, cx, cy, to_world=torch.eye(4), near=1e-6, far=1e7):
        super().__init__()

        self.add_param('fx', fx, help_msg='sensor focal length in x axis')
        self.add_param('fy', fy, help_msg='sensor focal length in y axis')
        self.add_param('cx', cx, help_msg='sensor principal point offset in x axis')
        self.add_param('cy', cy, help_msg='sensor principal point offset in y axis')
        self.add_param('near', near, help_msg='sensor near clip')
        self.add_param('far', far, help_msg='sensor far clip')
        self.add_param('to_world', to_torch_f(to_world), is_tensor=True, is_diff=True, help_msg='sensor to_world matrix')

    def get_rays(self, samples, aspect_ratio):
        samples = torch.cat([samples, torch.zeros_like(samples)[:, 0:1]], dim=1)
        sample_to_camera = torch.inverse(perspective_full(
            self['fx'], self['fy'], self['cx'], self['cy'], aspect_ratio, self['near'], self['far']))
        rays_o = batched_transform_pos(self['to_world'], to_torch_f([[0, 0, 0]]))
        rays_d = F.normalize(batched_transform_pos(sample_to_camera, samples), dim=1)
        rays_d = batched_transform_dir(self['to_world'], rays_d)
        return rays_o.repeat(samples.shape[0], 1), rays_d

    @classmethod
    def from_lookat(cls, fx, fy, cx, cy, origin, target, up, near=1e-6, far=1e7):
        sensor = cls(fx, fy, cx, cy)
        origin = to_torch_f(origin)
        target = to_torch_f(target)
        up = to_torch_f(up)
        sensor['to_world'] = lookat(origin, target, up)
        return sensor
        
class Mesh(ParamGroup):

    def __init__(self, v, f, uv, fuv, mat_id, to_world=torch.eye(4), use_face_normal=True, can_change_topology=False, radiance=torch.zeros(3)):
        super().__init__()
        
        self.add_param('v', to_torch_f(v), is_tensor=True, is_diff=True, help_msg='mesh vertex positions')
        self.add_param('f', to_torch_i(f), is_tensor=True, help_msg='mesh face indices')
        self.add_param('uv', to_torch_f(uv), is_tensor=True, help_msg='mesh uv coordinates')
        self.add_param('fuv', to_torch_i(fuv), is_tensor=True, help_msg='mesh uv face indices')
        self.add_param('mat_id', mat_id, help_msg='name of the material of the mesh')
        self.add_param('to_world', to_torch_f(to_world), is_tensor=True, help_msg='mesh to world matrix')
        self.add_param('use_face_normal', use_face_normal, help_msg='whether to use face normal')
        self.add_param('can_change_topology', can_change_topology, help_msg='whether to the topology can be chagned')

        radiance = to_torch_f(radiance)
        is_emitter = radiance.sum() > 0
        self.add_param('is_emitter', is_emitter, help_msg='whether it is used as an emitter')
        self.add_param('radiance', radiance, is_tensor=True, is_diff=True, help_msg='radiance if it is used as an emitter')

    @classmethod
    def from_file(cls, filename, mat_id, to_world=torch.eye(4), use_face_normal=True, can_change_topology=False, radiance=torch.zeros(3)):
        v, f, uv, fuv = read_mesh(filename)
        return cls(v, f, uv, fuv, mat_id, to_world, use_face_normal, can_change_topology, radiance)
    
class DiffuseBRDF(ParamGroup):

    def __init__(self, d):
        super().__init__()
        
        self.add_param('d', to_torch_f(d), is_tensor=True, is_diff=True, help_msg='diffuse reflectance')

    @classmethod
    def from_file(cls, filename, is_srgb=None):
        texture = read_image(filename, is_srgb)
        return cls(texture)
    
class MicrofacetBRDF(ParamGroup):

    def __init__(self, d, s, r):
        super().__init__()
        
        self.add_param('d', to_torch_f(d), is_tensor=True, is_diff=True, help_msg='diffuse reflectance')
        self.add_param('s', to_torch_f(s), is_tensor=True, is_diff=True, help_msg='specular reflectance')
        self.add_param('r', to_torch_f(r), is_tensor=True, is_diff=True, help_msg='roughness')

    @classmethod
    def from_file(cls, d_filename, s_filename, r_filename, d_is_srgb=None, s_is_srgb=None, r_is_srgb=None):
        d_texture = read_image(d_filename, d_is_srgb)
        s_texture = read_image(s_filename, s_is_srgb)
        r_texture = read_image(r_filename, r_is_srgb)[..., 0:1]

        return cls(d_texture, s_texture, r_texture)
    
class EnvironmentLight(ParamGroup):

    def __init__(self, radiance, to_world=torch.eye(4)):
        super().__init__()
        
        self.add_param('radiance', to_torch_f(radiance), is_tensor=True, is_diff=True, help_msg='environment light radiance')
        self.add_param('to_world', to_torch_f(to_world), is_tensor=True, is_diff=False, help_msg='environment to_world matrix')

    @classmethod
    def from_file(cls, radiance_filename, radiance_is_srgb=None, to_world=torch.eye(4)):
        radiance_texture = read_image(radiance_filename, radiance_is_srgb)

        return cls(radiance_texture, to_world)


from .config import *
import imageio
import imageio.v3 as iio
import numpy as np
import torch
from pathlib import Path
import gpytoolbox

# Download necessary imageio plugins. If they already exists they won't be 
# downloaded again. 
imageio.plugins.freeimage.download()

def read_mesh(mesh_path):
    v, f, uv, fuv = gpytoolbox.read_mesh(str(mesh_path), return_UV=True)
    if uv.size == 0:
        uv = np.zeros((0, 2))
        fuv = np.zeros((0, 3))
    return v, f, uv, fuv

def write_mesh(mesh_path, v, f, uv=None, fuv=None):
    v = to_numpy(v)
    f = to_numpy(f)
    if uv is not None: 
        uv = to_numpy(uv)
        if uv.size == 0:
            uv = None
    if fuv is not None: 
        fuv = to_numpy(fuv)
        if fuv.size == 0:
            fuv = None
    gpytoolbox.write_mesh(str(mesh_path), v, f, uv, fuv)

def linear_to_srgb(l):
    s = np.zeros_like(l)
    m = l <= 0.00313066844250063
    s[m] = l[m] * 12.92
    s[~m] = 1.055*(l[~m]**(1.0/2.4))-0.055
    return s

def srgb_to_linear(s):
    l = np.zeros_like(s)
    m = s <= 0.0404482362771082
    l[m] = s[m] / 12.92
    l[~m] = ((s[~m]+0.055)/1.055) ** 2.4
    return l

def to_srgb(image):
    image = to_numpy(image)
    if image.shape[2] == 4:
        image_alpha = image[:, :, 3:4]
        image = linear_to_srgb(image[:, :, 0:3])
        image = np.concatenate([image, image_alpha], axis=2)
    else:
        image = linear_to_srgb(image)
    return np.clip(image, 0, 1)

def to_linear(image):
    image = to_numpy(image)
    if image.shape[2] == 4:
        image_alpha = image[:, :, 3:4]
        image = srgb_to_linear(image[:, :, 0:3])
        image = np.concatenate([image, image_alpha], axis=2)
    else:
        image = srgb_to_linear(image)
    return image

def to_numpy(data):
    if torch.is_tensor(data):
        return data.detach().cpu().numpy()
    else:
        return np.array(data)
    
def to_torch(data, dtype):
    if torch.is_tensor(data):
        return data.to(dtype).to(device).contiguous()
    elif isinstance(data, np.ndarray):
        return torch.from_numpy(data).to(dtype).to(device).contiguous()
    else:
        return torch.tensor(data, dtype=dtype, device=device).contiguous()

def to_torch_f(data):
    return to_torch(data, ftype)

def to_torch_i(data):
    return to_torch(data, itype)

def read_image(image_path, is_srgb=None, remove_alpha=True):
    image_path = Path(image_path)
    
    image_ext = image_path.suffix
    iio_plugins = {
        '.exr': 'EXR-FI',
        '.hdr': 'HDR-FI',
        '.png': 'PNG-FI',
    }
    
    image = iio.imread(image_path, plugin=iio_plugins.get(image_ext))
    image = np.atleast_3d(image)
    
    if remove_alpha and image.shape[2] == 4:
        image = image[:, :, 0:3]

    if image.dtype == np.uint8 or image.dtype == np.int16:
        image = image.astype("float32") / 255.0
    elif image.dtype == np.uint16 or image.dtype == np.int32:
        image = image.astype("float32") / 65535.0

    if is_srgb is None:
        if image_ext in ['.exr', '.hdr', '.rgbe']:
            is_srgb = False
        else:
            is_srgb = True

    if is_srgb:
        image = to_linear(image)

    return image

def write_image(image_path, image, is_srgb=None):
    image_path = Path(image_path)
    
    image_ext = image_path.suffix
    iio_plugins = {
        '.exr': 'EXR-FI',
        '.hdr': 'HDR-FI',
        '.png': 'PNG-FI',
    }
    iio_flags = {
        '.exr': imageio.plugins.freeimage.IO_FLAGS.EXR_NONE,
    }
    hdr_formats = ['.exr', '.hdr', '.rgbe']
    
    image = to_numpy(image)
    image = np.atleast_3d(image)
    if image.shape[2] == 1:
        image = np.repeat(image, 3, axis=2)
        
    if image_ext in hdr_formats:
        is_srgb = False if is_srgb is None else is_srgb
    else:
        is_srgb = True if is_srgb is None else is_srgb
    if is_srgb:
        image = to_srgb(image)
        
    if image_ext in hdr_formats:
        image = image.astype(np.float32)
    else:
        image = (image * 255).astype(np.uint8)
    
    flags = iio_flags.get(image_ext)
    if flags is None: flags = 0
    
    iio.imwrite(image_path, image, 
                flags=flags,
                plugin=iio_plugins.get(image_ext))

def exr2png(image_path, verbose=False):
    image_path = Path(image_path)
    if image_path.is_dir():
        for p in image_path.glob('**/*.exr'):
            if verbose: print(p)
            im = read_image(p)
            write_image(p.with_suffix('.png'), im)
    elif image_path.suffix == '.exr':
        if verbose: print(image_path)
        im = read_image(image_path)
        write_image(image_path.with_suffix('.png'), im)

def write_video(video_path, frames, fps=20, kwargs={}):
    video_path = Path(video_path)
    video_path.parent.mkdir(exist_ok=True, parents=True)
    writer = imageio.get_writer(video_path, fps=fps, **kwargs)
    for frame in frames:
        frame = (to_srgb(to_numpy(frame)) * 255).astype(np.uint8)
        writer.append_data(frame)
    writer.close()


"""Inverse-Rendering Toolkit"""

__version__ = "0.1.2"

from .connector import is_connector_available, get_connector_list, get_connector

try:
    from .connectors.psdr_jit_connector import PSDRJITConnector
except:
    pass

from .renderer import Renderer


from .connector import get_connector
from .config import *
import torch 
import numpy as np

import gin

class RenderFunction(torch.autograd.Function):

    @staticmethod
    def forward(ctx, connector, scene, render_options, sensor_ids, integrator_id, *params):
        images = connector.renderC(scene, render_options, sensor_ids=sensor_ids, integrator_id=integrator_id)
        images = torch.stack(images, dim=0)

        ctx.connector = connector
        ctx.scene = scene
        ctx.render_options = render_options
        ctx.sensor_ids = sensor_ids
        ctx.integrator_id = integrator_id
        ctx.num_no_grads = 5 # number of inputs that don't require grad

        ctx.params = params
       
        assert(images.sum().isfinite())
        return images

    @staticmethod
    def backward(ctx, grad_out):
        image_grads = [image_grad for image_grad in grad_out]
        param_grads = ctx.connector.renderD(image_grads, ctx.scene, ctx.render_options, ctx.sensor_ids, ctx.integrator_id)
        return tuple([None] * ctx.num_no_grads + param_grads)

@gin.configurable
class Renderer(torch.nn.Module):

    def __init__(self, connector_name, render_options=None):
        super().__init__()
        self.connector = get_connector(connector_name)
        self.render_options = render_options

    def forward(self, scene, sensor_ids=[0], integrator_id=0):
        assert self.render_options is not None, "Please set render options first."
        if torch.is_tensor(sensor_ids):
            sensor_ids = sensor_ids.flatten().tolist()

        params = [scene[param_name] for param_name in scene.requiring_grad]
        
        images = RenderFunction.apply(self.connector, scene, self.render_options, sensor_ids, integrator_id, *params)
        return images

    def set_render_options(self, render_options):
        self.render_options = render_options


import trimesh
from chamferdist import ChamferDistance
import numpy as np
import torch

def chamfer_distance(mesh_a_path, mesh_b_path, num_samples, mode='bidirectional'):
    assert mode in ['forward', 'backward', 'bidirectional']

    mesh_a = trimesh.load_mesh(mesh_a_path)
    mesh_b = trimesh.load_mesh(mesh_b_path)
    samples_a, _ = trimesh.sample.sample_surface(mesh_a, num_samples)
    samples_b, _ = trimesh.sample.sample_surface(mesh_b, num_samples)
    samples_a = torch.from_numpy(samples_a.view(np.ndarray)).unsqueeze(0).cuda()
    samples_b = torch.from_numpy(samples_b.view(np.ndarray)).unsqueeze(0).cuda()

    reverse = False
    bidirectional = False

    if mode == 'backward':
        reverse = True
    elif mode == 'bidirectional':
        bidirectional = True

    cd = ChamferDistance()
    dist = cd(samples_a, samples_b, 
              reverse=reverse, 
              bidirectional=bidirectional).item()
    mean_dist = dist / num_samples

    return mean_dist


from .io import to_torch
import torch
from collections import OrderedDict

class ParamGroup:

    def __init__(self):
        self.params = OrderedDict()

    def __getitem__(self, param_name):
        return self.params[param_name]['value']
    
    def __setitem__(self, param_name, param_value):
        if self.params[param_name]['is_tensor'] and not torch.is_tensor(param_value):
            dtype = self[param_name].dtype
            requires_grad = self[param_name].requires_grad
            param_value = to_torch(param_value, dtype).requires_grad_(requires_grad)
        self.params[param_name]['value'] = param_value
        self.mark_updated(param_name)

    def add_param(self, name, value, is_tensor=False, is_diff=False, help_msg=""):
        self.params[name] = {
            'value': value,
            'is_tensor': is_tensor,
            'is_diff': is_diff,
            'help_msg': help_msg,
            'updated': False
        }

    def get_requiring_grad(self):
        return [name for name in self.params if self.params[name]['is_diff'] and self[name].requires_grad]
    
    def get_updated(self):
        return [name for name in self.params if self.params[name]['updated']]
    
    def mark_updated(self, param_name):
        self.params[param_name]['updated'] = True
    
    def __str__(self):
        lines = []

        lines.append('----')

        for name in self.params:
            lines.append(f"{name}:")
            lines.append(f"\tdescription: {self.params[name]['help_msg']}")

            is_diff = self.params[name]['is_diff']
            lines.append(f"\tis differentiable: {is_diff}")
            if is_diff:
                lines.append(f"\trequires grad: {self[name].requires_grad}")

            lines.append(f"\tvalue: \t{self[name]}")

        lines.append('----')

        return '\n'.join(lines)




import torch
from abc import ABC, abstractmethod
from pathlib import Path
from .config import *

import gin 

class Model(ABC):
    def __init__(self, scene):
        self.scene = scene

    @abstractmethod
    def zero_grad(self):
        pass

    @abstractmethod
    def set_data(self):
        pass

    @abstractmethod
    def step(self):
        pass

    @abstractmethod
    def get_results(self):
        pass

    @abstractmethod
    def write_results(self, result_path):
        pass

    def get_regularization(self):
        return torch.tensor(0.0, device=device, dtype=ftype)

@gin.configurable
class MultiOpt(Model):

    def __init__(self, scene, model_classes):
        super().__init__(scene)

        self._models = [model_class(scene) for model_class in model_classes]

    def zero_grad(self):
        for model in self._models:
            model.zero_grad()

    def set_data(self):
        for model in self._models:
            model.set_data()
        
    def step(self):
        for model in self._models:
            model.step()

    def get_results(self):
        results = []
        for model in self._models:
            results.append(model.get_results())
        return results

    def write_results(self, result_path):
        for model in self._models:
            model.write_results(result_path)

    def get_regularization(self):
        reg = torch.tensor(0.0, device=device, dtype=ftype)
        for model in self._models:
            reg += model.get_regularization()
        return reg